@@ -28,7 +28,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
-#include <getopt.h>
 #include <stdbool.h>
 
 #include <glib.h>
@@ -38,9 +37,17 @@
 #include "input.h"
 #include "services.h"
 #include "commands.h"
+#include "agent.h"
+#include "vpnconnections.h"
 
 static DBusConnection *connection;
 
+struct connman_option {
+	const char *name;
+	const char val;
+	const char *desc;
+};
+
 static char *ipv4[] = {
 	"Method",
 	"Address",
@@ -57,7 +64,28 @@ static char *ipv6[] = {
 	NULL
 };
 
-static int cmd_help(char *args[], int num, struct option *options);
+static int cmd_help(char *args[], int num, struct connman_option *options);
+
+static bool check_dbus_name(const char *name)
+{
+	/*
+	 * Valid dbus chars should be [A-Z][a-z][0-9]_
+	 * and should not start with number.
+	 */
+	unsigned int i;
+
+	if (name == NULL || name[0] == '\0')
+		return false;
+
+	for (i = 0; name[i] != '\0'; i++)
+		if (!((name[i] >= 'A' && name[i] <= 'Z') ||
+				(name[i] >= 'a' && name[i] <= 'z') ||
+				(name[i] >= '0' && name[i] <= '9') ||
+				name[i] == '_'))
+			return false;
+
+	return true;
+}
 
 static int parse_boolean(char *arg)
 {
@@ -85,7 +113,7 @@ static int parse_boolean(char *arg)
 	return -1;
 }
 
-static int parse_args(char *arg, struct option *options)
+static int parse_args(char *arg, struct connman_option *options)
 {
 	int i;
 
@@ -102,7 +130,7 @@ static int parse_args(char *arg, struct option *options)
 	return '?';
 }
 
-static void enable_return(DBusMessageIter *iter, const char *error,
+static int enable_return(DBusMessageIter *iter, const char *error,
 		void *user_data)
 {
 	char *tech = user_data;
@@ -120,9 +148,11 @@ static void enable_return(DBusMessageIter *iter, const char *error,
 		fprintf(stderr, "Error %s: %s\n", str, error);
 
 	g_free(user_data);
+
+	return 0;
 }
 
-static int cmd_enable(char *args[], int num, struct option *options)
+static int cmd_enable(char *args[], int num, struct connman_option *options)
 {
 	char *tech;
 	dbus_bool_t b = TRUE;
@@ -133,6 +163,9 @@ static int cmd_enable(char *args[], int num, struct option *options)
 	if (num < 2)
 		return -EINVAL;
 
+	if (check_dbus_name(args[1]) == false)
+		return -EINVAL;
+
 	if (strcmp(args[1], "offlinemode") == 0) {
 		tech = g_strdup(args[1]);
 		return __connmanctl_dbus_set_property(connection, "/",
@@ -146,7 +179,7 @@ static int cmd_enable(char *args[], int num, struct option *options)
 				"Powered", DBUS_TYPE_BOOLEAN, &b);
 }
 
-static void disable_return(DBusMessageIter *iter, const char *error,
+static int disable_return(DBusMessageIter *iter, const char *error,
 		void *user_data)
 {
 	char *tech = user_data;
@@ -164,9 +197,11 @@ static void disable_return(DBusMessageIter *iter, const char *error,
 		fprintf(stderr, "Error %s: %s\n", str, error);
 
 	g_free(user_data);
+
+	return 0;
 }
 
-static int cmd_disable(char *args[], int num, struct option *options)
+static int cmd_disable(char *args[], int num, struct connman_option *options)
 {
 	char *tech;
 	dbus_bool_t b = FALSE;
@@ -177,6 +212,9 @@ static int cmd_disable(char *args[], int num, struct option *options)
 	if (num < 2)
 		return -EINVAL;
 
+	if (check_dbus_name(args[1]) == false)
+		return -EINVAL;
+
 	if (strcmp(args[1], "offlinemode") == 0) {
 		tech = g_strdup(args[1]);
 		return __connmanctl_dbus_set_property(connection, "/",
@@ -190,32 +228,34 @@ static int cmd_disable(char *args[], int num, struct option *options)
 				"Powered", DBUS_TYPE_BOOLEAN, &b);
 }
 
-static void state_print(DBusMessageIter *iter, const char *error,
+static int state_print(DBusMessageIter *iter, const char *error,
 		void *user_data)
 {
 	DBusMessageIter entry;
 
 	if (error != NULL) {
 		fprintf(stderr, "Error: %s", error);
-		return;
+		return 0;
 	}
 
 	dbus_message_iter_recurse(iter, &entry);
 	__connmanctl_dbus_print(&entry, "  ", " = ", "\n");
 	fprintf(stdout, "\n");
+
+	return 0;
 }
 
-static int cmd_state(char *args[], int num, struct option *options)
+static int cmd_state(char *args[], int num, struct connman_option *options)
 {
 	if (num > 1)
 		return -E2BIG;
 
-	return __connmanctl_dbus_method_call(connection, "/",
-			"net.connman.Manager", "GetProperties",
+	return __connmanctl_dbus_method_call(connection, CONNMAN_SERVICE,
+			CONNMAN_PATH, "net.connman.Manager", "GetProperties",
 			state_print, NULL, DBUS_TYPE_INVALID);
 }
 
-static void services_list(DBusMessageIter *iter, const char *error,
+static int services_list(DBusMessageIter *iter, const char *error,
 		void *user_data)
 {
 	if (error == NULL) {
@@ -224,9 +264,11 @@ static void services_list(DBusMessageIter *iter, const char *error,
 	} else {
 		fprintf(stderr, "Error: %s\n", error);
 	}
+
+	return 0;
 }
 
-static void services_properties(DBusMessageIter *iter, const char *error,
+static int services_properties(DBusMessageIter *iter, const char *error,
 		void *user_data)
 {
 	char *path = user_data;
@@ -252,9 +294,11 @@ static void services_properties(DBusMessageIter *iter, const char *error,
 	}
 
 	g_free(user_data);
+
+	return 0;
 }
 
-static int cmd_services(char *args[], int num, struct option *options)
+static int cmd_services(char *args[], int num, struct connman_option *options)
 {
 	char *service_name = NULL;
 	char *path;
@@ -280,25 +324,29 @@ static int cmd_services(char *args[], int num, struct option *options)
 	}
 
 	if (service_name == NULL) {
-		return __connmanctl_dbus_method_call(connection, "/",
-			"net.connman.Manager", "GetServices",
-			services_list, NULL, DBUS_TYPE_INVALID);
+		return __connmanctl_dbus_method_call(connection,
+				CONNMAN_SERVICE, CONNMAN_PATH,
+				"net.connman.Manager", "GetServices",
+				services_list, NULL, DBUS_TYPE_INVALID);
 	}
 
+	if (check_dbus_name(service_name) == false)
+		return -EINVAL;
+
 	path = g_strdup_printf("/net/connman/service/%s", service_name);
-	return __connmanctl_dbus_method_call(connection, path,
+	return __connmanctl_dbus_method_call(connection, CONNMAN_SERVICE, path,
 			"net.connman.Service", "GetProperties",
 			services_properties, path, DBUS_TYPE_INVALID);
 }
 
-static void technology_print(DBusMessageIter *iter, const char *error,
+static int technology_print(DBusMessageIter *iter, const char *error,
 		void *user_data)
 {
 	DBusMessageIter array;
 
 	if (error != NULL) {
 		fprintf(stderr, "Error: %s\n", error);
-		return;
+		return 0;
 	}
 
 	dbus_message_iter_recurse(iter, &array);
@@ -318,19 +366,201 @@ static void technology_print(DBusMessageIter *iter, const char *error,
 
 		dbus_message_iter_next(&array);
 	}
+
+	return 0;
 }
 
-static int cmd_technologies(char *args[], int num, struct option *options)
+static int cmd_technologies(char *args[], int num,
+		struct connman_option *options)
 {
 	if (num > 1)
 		return -E2BIG;
 
-	return __connmanctl_dbus_method_call(connection, "/",
-			"net.connman.Manager", "GetTechnologies",
+	return __connmanctl_dbus_method_call(connection, CONNMAN_SERVICE,
+			CONNMAN_PATH, "net.connman.Manager", "GetTechnologies",
 			technology_print, NULL,	DBUS_TYPE_INVALID);
 }
 
-static void scan_return(DBusMessageIter *iter, const char *error,
+struct tether_enable {
+	char *path;
+	dbus_bool_t enable;
+};
+
+static int tether_set_return(DBusMessageIter *iter, const char *error,
+		void *user_data)
+{
+	struct tether_enable *tether = user_data;
+	char *str;
+
+	str = strrchr(tether->path, '/');
+	if (str != NULL)
+		str++;
+	else
+		str = tether->path;
+
+	if (error == NULL) {
+		fprintf(stdout, "%s tethering for %s\n",
+				tether->enable == TRUE ? "Enabled": "Disabled",
+				str);
+	} else
+		fprintf(stderr, "Error %s %s tethering: %s\n",
+				tether->enable == TRUE ?
+				"enabling": "disabling", str, error);
+
+	g_free(tether->path);
+	g_free(user_data);
+
+	return 0;
+}
+
+static int tether_set(char *technology, int set_tethering)
+{
+	struct tether_enable *tether = g_new(struct tether_enable, 1);
+
+	switch(set_tethering) {
+	case 1:
+		tether->enable = TRUE;
+		break;
+	case 0:
+		tether->enable = FALSE;
+		break;
+	default:
+		g_free(tether);
+		return 0;
+	}
+
+	tether->path = g_strdup_printf("/net/connman/technology/%s",
+			technology);
+
+	return __connmanctl_dbus_set_property(connection, tether->path,
+			"net.connman.Technology", tether_set_return,
+			tether, "Tethering", DBUS_TYPE_BOOLEAN,
+			&tether->enable);
+}
+
+struct tether_properties {
+	int ssid_result;
+	int passphrase_result;
+	int set_tethering;
+};
+
+static int tether_update(struct tether_properties *tether)
+{
+	printf("%d %d %d\n", tether->ssid_result, tether->passphrase_result,
+		tether->set_tethering);
+
+	if (tether->ssid_result == 0 && tether->passphrase_result == 0)
+		return tether_set("wifi", tether->set_tethering);
+
+	if (tether->ssid_result != -EINPROGRESS &&
+			tether->passphrase_result != -EINPROGRESS) {
+		g_free(tether);
+		return 0;
+	}
+
+	return -EINPROGRESS;
+}
+
+static int tether_set_ssid_return(DBusMessageIter *iter, const char *error,
+		void *user_data)
+{
+	struct tether_properties *tether = user_data;
+
+	if (error == NULL) {
+		fprintf(stdout, "Wifi SSID set\n");
+		tether->ssid_result = 0;
+	} else {
+		fprintf(stderr, "Error setting wifi SSID: %s\n", error);
+		tether->ssid_result = -EINVAL;
+	}
+
+	return tether_update(tether);
+}
+
+static int tether_set_passphrase_return(DBusMessageIter *iter,
+		const char *error, void *user_data)
+{
+	struct tether_properties *tether = user_data;
+
+	if (error == NULL) {
+		fprintf(stdout, "Wifi passphrase set\n");
+		tether->passphrase_result = 0;
+	} else {
+		fprintf(stderr, "Error setting wifi passphrase: %s\n", error);
+		tether->passphrase_result = -EINVAL;
+	}
+
+	return tether_update(tether);
+}
+
+static int tether_set_ssid(char *ssid, char *passphrase, int set_tethering)
+{
+	struct tether_properties *tether = g_new(struct tether_properties, 1);
+
+	tether->set_tethering = set_tethering;
+
+	tether->ssid_result = __connmanctl_dbus_set_property(connection,
+			"/net/connman/technology/wifi",
+			"net.connman.Technology",
+			tether_set_ssid_return, tether,
+			"TetheringIdentifier", DBUS_TYPE_STRING, &ssid);
+
+	tether->passphrase_result =__connmanctl_dbus_set_property(connection,
+			"/net/connman/technology/wifi",
+			"net.connman.Technology",
+			tether_set_passphrase_return, tether,
+			"TetheringPassphrase", DBUS_TYPE_STRING, &passphrase);
+
+	if (tether->ssid_result != -EINPROGRESS &&
+			tether->passphrase_result != -EINPROGRESS) {
+		g_free(tether);
+		return -ENXIO;
+	}
+
+	return -EINPROGRESS;
+}
+
+static int cmd_tether(char *args[], int num, struct connman_option *options)
+{
+	char *ssid, *passphrase;
+	int set_tethering;
+
+	if (num < 3)
+		return -EINVAL;
+
+	passphrase = args[num - 1];
+	ssid = args[num - 2];
+
+	set_tethering = parse_boolean(args[2]);
+
+	if (strcmp(args[1], "wifi") == 0) {
+
+		if (num > 5)
+			return -E2BIG;
+
+		if (num == 5 && set_tethering == -1)
+			return -EINVAL;
+
+		if (num == 4)
+			set_tethering = -1;
+
+		if (num > 3)
+			return tether_set_ssid(ssid, passphrase, set_tethering);
+	}
+
+	if (num > 3)
+		return -E2BIG;
+
+	if (set_tethering == -1)
+		return -EINVAL;
+
+	if (check_dbus_name(args[1]) == false)
+		return -EINVAL;
+
+	return tether_set(args[1], set_tethering);
+}
+
+static int scan_return(DBusMessageIter *iter, const char *error,
 		void *user_data)
 {
 	char *path = user_data;
@@ -340,12 +570,14 @@ static void scan_return(DBusMessageIter *iter, const char *error,
 		str++;
 		fprintf(stdout, "Scan completed for %s\n", str);
 	} else
-		fprintf(stderr, "Error %s: %s", path, error);
+		fprintf(stderr, "Error %s: %s\n", path, error);
 
 	g_free(user_data);
+
+	return 0;
 }
 
-static int cmd_scan(char *args[], int num, struct option *options)
+static int cmd_scan(char *args[], int num, struct connman_option *options)
 {
 	char *path;
 
@@ -355,13 +587,16 @@ static int cmd_scan(char *args[], int num, struct option *options)
 	if (num < 2)
 		return -EINVAL;
 
+	if (check_dbus_name(args[1]) == false)
+		return -EINVAL;
+
 	path = g_strdup_printf("/net/connman/technology/%s", args[1]);
-	return __connmanctl_dbus_method_call(connection, path,
+	return __connmanctl_dbus_method_call(connection, CONNMAN_SERVICE, path,
 			"net.connman.Technology", "Scan",
 			scan_return, path, DBUS_TYPE_INVALID);
 }
 
-static void connect_return(DBusMessageIter *iter, const char *error,
+static int connect_return(DBusMessageIter *iter, const char *error,
 		void *user_data)
 {
 	char *path = user_data;
@@ -374,9 +609,11 @@ static void connect_return(DBusMessageIter *iter, const char *error,
 		fprintf(stderr, "Error %s: %s\n", path, error);
 
 	g_free(user_data);
+
+	return 0;
 }
 
-static int cmd_connect(char *args[], int num, struct option *options)
+static int cmd_connect(char *args[], int num, struct connman_option *options)
 {
 	char *path;
 
@@ -386,13 +623,16 @@ static int cmd_connect(char *args[], int num, struct option *options)
 	if (num < 2)
 		return -EINVAL;
 
+	if (check_dbus_name(args[1]) == false)
+		return -EINVAL;
+
 	path = g_strdup_printf("/net/connman/service/%s", args[1]);
-	return __connmanctl_dbus_method_call(connection, path,
+	return __connmanctl_dbus_method_call(connection, CONNMAN_SERVICE, path,
 			"net.connman.Service", "Connect",
 			connect_return, path, DBUS_TYPE_INVALID);
 }
 
-static void disconnect_return(DBusMessageIter *iter, const char *error,
+static int disconnect_return(DBusMessageIter *iter, const char *error,
 		void *user_data)
 {
 	char *path = user_data;
@@ -405,9 +645,11 @@ static void disconnect_return(DBusMessageIter *iter, const char *error,
 		fprintf(stderr, "Error %s: %s\n", path, error);
 
 	g_free(user_data);
+
+	return 0;
 }
 
-static int cmd_disconnect(char *args[], int num, struct option *options)
+static int cmd_disconnect(char *args[], int num, struct connman_option *options)
 {
 	char *path;
 
@@ -417,15 +659,16 @@ static int cmd_disconnect(char *args[], int num, struct option *options)
 	if (num < 2)
 		return -EINVAL;
 
+	if (check_dbus_name(args[1]) == false)
+		return -EINVAL;
+
 	path = g_strdup_printf("/net/connman/service/%s", args[1]);
-	return __connmanctl_dbus_method_call(connection, path,
+	return __connmanctl_dbus_method_call(connection, CONNMAN_SERVICE, path,
 			"net.connman.Service", "Disconnect",
 			disconnect_return, path, DBUS_TYPE_INVALID);
-
-	return 0;
 }
 
-static void config_return(DBusMessageIter *iter, const char *error,
+static int config_return(DBusMessageIter *iter, const char *error,
 		void *user_data)
 {
 	char *service_name = user_data;
@@ -434,6 +677,8 @@ static void config_return(DBusMessageIter *iter, const char *error,
 		fprintf(stderr, "Error %s: %s\n", service_name, error);
 
 	g_free(user_data);
+
+	return 0;
 }
 
 struct config_append {
@@ -494,10 +739,12 @@ static void config_append_ipv6(DBusMessageIter *iter, void *user_data)
 			if (opts[1] != NULL) {
 				append->values = 2;
 
-				if (g_strcmp0(opts[0], "prefered") != 0) {
+				if (g_strcmp0(opts[1], "prefered") != 0 &&
+						g_strcmp0(opts[1],
+							"preferred") != 0) {
 					fprintf(stderr, "Error %s: %s\n",
 							opts[1],
-							strerror(-EINVAL));
+							strerror(EINVAL));
 					return;
 				}
 
@@ -624,7 +871,7 @@ static void config_append_proxy(DBusMessageIter *iter, void *user_data)
 	append->values++;
 }
 
-static int cmd_config(char *args[], int num, struct option *options)
+static int cmd_config(char *args[], int num, struct connman_option *options)
 {
 	int result = 0, res = 0, index = 2, oldindex = 0;
 	int c;
@@ -637,6 +884,9 @@ static int cmd_config(char *args[], int num, struct option *options)
 	if (service_name == NULL)
 		return -EINVAL;
 
+	if (check_dbus_name(service_name) == false)
+		return -EINVAL;
+
 	while (index < num && args[index] != NULL) {
 		c = parse_args(args[index], options);
 		opt_start = &args[index + 1];
@@ -732,7 +982,8 @@ static int cmd_config(char *args[], int num, struct option *options)
 			break;
 		case 'r':
 			res = __connmanctl_dbus_method_call(connection,
-					path, "net.connman.Service", "Remove",
+					CONNMAN_SERVICE, path,
+					"net.connman.Service", "Remove",
 					config_return, g_strdup(service_name),
 					DBUS_TYPE_INVALID);
 			break;
@@ -768,6 +1019,10 @@ static DBusHandlerResult monitor_changed(DBusConnection *connection,
 	if (strncmp(interface, "net.connman.", 12) != 0)
 		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 
+	if (strncmp(interface, "net.connman.Agent", 17) == 0 ||
+			strncmp(interface, "net.connman.vpn.Agent", 21) == 0)
+		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
 	interface = strrchr(interface, '.');
 	if (interface != NULL && *interface != '\0')
 		interface++;
@@ -792,13 +1047,20 @@ static DBusHandlerResult monitor_changed(DBusConnection *connection,
 		return DBUS_HANDLER_RESULT_HANDLED;
 	}
 
-	if (dbus_message_is_signal(message, "net.connman.Manager",
-					"TechnologyAdded") == TRUE)
-		path = "TechnologyAdded";
 
-	if (dbus_message_is_signal(message, "net.connman.Manager",
+	if (dbus_message_is_signal(message, "net.connman.vpn.Manager",
+					"ConnectionAdded") == TRUE ||
+			dbus_message_is_signal(message,
+					"net.connman.vpn.Manager",
+					"ConnectionRemoved") == TRUE) {
+		interface = "vpn.Manager";
+		path = dbus_message_get_member(message);
+
+	} else if (dbus_message_is_signal(message, "net.connman.Manager",
+					"TechnologyAdded") == TRUE ||
+			dbus_message_is_signal(message, "net.connman.Manager",
 					"TechnologyRemoved") == TRUE)
-		path = "TechnologyRemoved";
+		path = dbus_message_get_member(message);
 
 	fprintf(stdout, "%-12s %-20s ", interface, path);
 	dbus_message_iter_init(message, &iter);
@@ -811,34 +1073,45 @@ static DBusHandlerResult monitor_changed(DBusConnection *connection,
 	return DBUS_HANDLER_RESULT_HANDLED;
 }
 
-static bool monitor_s = false;
-static bool monitor_t = false;
-static bool monitor_m = false;
+static struct {
+	char *interface;
+	bool enabled;
+} monitor[] = {
+	{ "Service", false },
+	{ "Technology", false },
+	{ "Manager", false },
+	{ "vpn.Manager", false },
+	{ "vpn.Connection", false },
+	{ NULL, },
+};
 
 static void monitor_add(char *interface)
 {
+	bool add_filter = true, found = false;
+	int i;
 	char *rule;
 	DBusError err;
 
-	if (monitor_s == false && monitor_t == false && monitor_m == false)
+	for (i = 0; monitor[i].interface != NULL; i++) {
+		if (monitor[i].enabled == true)
+			add_filter = false;
+
+		if (g_strcmp0(interface, monitor[i].interface) == 0) {
+			if (monitor[i].enabled == true)
+				return;
+
+			monitor[i].enabled = true;
+			found = true;
+		}
+	}
+
+	if (found == false)
+		return;
+
+	if (add_filter == true)
 		dbus_connection_add_filter(connection, monitor_changed,
 				NULL, NULL);
 
-	if (g_strcmp0(interface, "Service") == 0) {
-		if (monitor_s == true)
-			return;
-		monitor_s = true;
-	} else if (g_strcmp0(interface, "Technology") == 0) {
-		if (monitor_t == true)
-			return;
-		monitor_t = true;
-	} else if (g_strcmp0(interface, "Manager") == 0) {
-		if (monitor_m == true)
-			return;
-		monitor_m = true;
-	} else
-		return;
-
 	dbus_error_init(&err);
 	rule  = g_strdup_printf("type='signal',interface='net.connman.%s'",
 			interface);
@@ -851,21 +1124,25 @@ static void monitor_add(char *interface)
 
 static void monitor_del(char *interface)
 {
+	bool del_filter = true, found = false;
+	int i;
 	char *rule;
 
-	if (g_strcmp0(interface, "Service") == 0) {
-		if (monitor_s == false)
-			return;
-		monitor_s = false;
-	} else if (g_strcmp0(interface, "Technology") == 0) {
-		if (monitor_t == false)
-			return;
-		monitor_t = false;
-	} else if (g_strcmp0(interface, "Manager") == 0) {
-		if (monitor_m == false)
-			return;
-		monitor_m = false;
-	} else
+
+	for (i = 0; monitor[i].interface != NULL; i++) {
+		if (g_strcmp0(interface, monitor[i].interface) == 0) {
+			if (monitor[i].enabled == false)
+				return;
+
+			monitor[i].enabled = false;
+			found = true;
+		}
+
+		if (monitor[i].enabled == true)
+			del_filter = false;
+	}
+
+	if (found == false)
 		return;
 
 	rule  = g_strdup_printf("type='signal',interface='net.connman.%s'",
@@ -873,12 +1150,12 @@ static void monitor_del(char *interface)
 	dbus_bus_remove_match(connection, rule, NULL);
 	g_free(rule);
 
-	if (monitor_s == false && monitor_t == false && monitor_m == false)
+	if (del_filter == true)
 		dbus_connection_remove_filter(connection, monitor_changed,
 				NULL);
 }
 
-static int cmd_monitor(char *args[], int num, struct option *options)
+static int cmd_monitor(char *args[], int num, struct connman_option *options)
 {
 	bool add = true;
 	int c;
@@ -903,6 +1180,8 @@ static int cmd_monitor(char *args[], int num, struct option *options)
 		monitor_add("Service");
 		monitor_add("Technology");
 		monitor_add("Manager");
+		monitor_add("vpn.Manager");
+		monitor_add("vpn.Connection");
 		break;
 
 	case 's':
@@ -926,18 +1205,36 @@ static int cmd_monitor(char *args[], int num, struct option *options)
 			monitor_del("Manager");
 		break;
 
+	case 'M':
+		if (add == true)
+			monitor_add("vpn.Manager");
+		else
+			monitor_del("vpn.Manager");
+		break;
+
+	case 'C':
+		if (add == true)
+			monitor_add("vpn.Connection");
+		else
+			monitor_del("vpn.Connection");
+		break;
+
 	default:
 		switch(parse_boolean(args[1])) {
 		case 0:
 			monitor_del("Service");
 			monitor_del("Technology");
 			monitor_del("Manager");
+			monitor_del("vpn.Manager");
+			monitor_del("vpn.Connection");
 			break;
 
 		case 1:
 			monitor_add("Service");
 			monitor_add("Technology");
 			monitor_add("Manager");
+			monitor_add("vpn.Manager");
+			monitor_add("vpn.Connection");
 			break;
 
 		default:
@@ -951,99 +1248,212 @@ static int cmd_monitor(char *args[], int num, struct option *options)
 	return 0;
 }
 
-static int cmd_exit(char *args[], int num, struct option *options)
+static int cmd_agent(char *args[], int num, struct connman_option *options)
+{
+	if (num > 2)
+		return -E2BIG;
+
+	if (num < 2)
+		return -EINVAL;
+
+	switch(parse_boolean(args[1])) {
+	case 0:
+		__connmanctl_agent_unregister(connection);
+		break;
+
+	case 1:
+		if (__connmanctl_agent_register(connection) == -EINPROGRESS)
+			return -EINPROGRESS;
+
+		break;
+
+	default:
+		return -EINVAL;
+		break;
+	}
+
+	return 0;
+}
+
+static int vpnconnections_properties(DBusMessageIter *iter, const char *error,
+		void *user_data)
+{
+	char *path = user_data;
+	char *str;
+	DBusMessageIter dict;
+
+	if (error == NULL) {
+		fprintf(stdout, "%s\n", path);
+
+		dbus_message_iter_recurse(iter, &dict);
+		__connmanctl_dbus_print(&dict, "  ", " = ", "\n");
+
+		fprintf(stdout, "\n");
+
+	} else {
+		str = strrchr(path, '/');
+		if (str != NULL)
+			str++;
+		else
+			str = path;
+
+		fprintf(stderr, "Error %s: %s\n", str, error);
+	}
+
+	g_free(user_data);
+
+	return 0;
+}
+
+static int vpnconnections_list(DBusMessageIter *iter, const char *error,
+		void *user_data)
+{
+	if (error == NULL)
+		__connmanctl_vpnconnections_list(iter);
+        else
+		fprintf(stderr, "Error: %s\n", error);
+
+	return 0;
+}
+
+static int cmd_vpnconnections(char *args[], int num,
+		struct connman_option *options)
+{
+	char *vpnconnection_name, *path;
+
+	if (num > 2)
+		return -E2BIG;
+
+	vpnconnection_name = args[1];
+
+	if (vpnconnection_name == NULL)
+		return __connmanctl_dbus_method_call(connection,
+				VPN_SERVICE, VPN_PATH,
+				"net.connman.vpn.Manager", "GetConnections",
+				vpnconnections_list, NULL,
+				DBUS_TYPE_INVALID);
+
+	if (check_dbus_name(vpnconnection_name) == false)
+		return -EINVAL;
+
+	path = g_strdup_printf("/net/connman/vpn/connection/%s",
+			vpnconnection_name);
+	return __connmanctl_dbus_method_call(connection, VPN_SERVICE, path,
+			"net.connman.vpn.Connection", "GetProperties",
+			vpnconnections_properties, path, DBUS_TYPE_INVALID);
+
+}
+
+static int cmd_vpnagent(char *args[], int num, struct connman_option *options)
+{
+	if (num > 2)
+		return -E2BIG;
+
+	if (num < 2)
+		return -EINVAL;
+
+	switch(parse_boolean(args[1])) {
+	case 0:
+		__connmanctl_vpn_agent_unregister(connection);
+		break;
+
+	case 1:
+		if (__connmanctl_vpn_agent_register(connection) ==
+				-EINPROGRESS)
+			return -EINPROGRESS;
+
+		break;
+
+	default:
+		return -EINVAL;
+		break;
+	}
+
+	return 0;
+}
+
+static int cmd_exit(char *args[], int num, struct connman_option *options)
 {
 	return 1;
 }
 
-static struct option service_options[] = {
-	{"properties", required_argument, 0, 'p'},
+static struct connman_option service_options[] = {
+	{"properties", 'p', "[<service>]      (obsolete)"},
 	{ NULL, }
 };
 
-static const char *service_desc[] = {
-	"[<service>]      (obsolete)",
-	NULL
-};
-
-static struct option config_options[] = {
-	{"nameservers", required_argument, 0, 'n'},
-	{"timeservers", required_argument, 0, 't'},
-	{"domains", required_argument, 0, 'd'},
-	{"ipv6", required_argument, 0, 'v'},
-	{"proxy", required_argument, 0, 'x'},
-	{"autoconnect", required_argument, 0, 'a'},
-	{"ipv4", required_argument, 0, 'i'},
-	{"remove", 0, 0, 'r'},
+static struct connman_option config_options[] = {
+	{"nameservers", 'n', "<dns1> [<dns2>] [<dns3>]"},
+	{"timeservers", 't', "<ntp1> [<ntp2>] [...]"},
+	{"domains", 'd', "<domain1> [<domain2>] [...]"},
+	{"ipv6", 'v', "off|auto [enable|disable|prefered]|\n"
+	              "\t\t\tmanual <address> <prefixlength> <gateway>"},
+	{"proxy", 'x', "direct|auto <URL>|manual <URL1> [<URL2>] [...]\n"
+	               "\t\t\t[exclude <exclude1> [<exclude2>] [...]]"},
+	{"autoconnect", 'a', "yes|no"},
+	{"ipv4", 'i', "off|dhcp|manual <address> <netmask> <gateway>"},
+	{"remove", 'r', "                 Remove service"},
 	{ NULL, }
 };
 
-static const char *config_desc[] = {
-	"<dns1> [<dns2>] [<dns3>]",
-	"<ntp1> [<ntp2>] [...]",
-	"<domain1> [<domain2>] [...]",
-	"off|auto [enable|disable|prefered]|\n"
-	"\t\t\tmanual <address> <prefixlength> <gateway>",
-	"direct|auto <URL>|manual <URL1> [<URL2>] [...]\n"
-	"\t\t\t[exclude <exclude1> [<exclude2>] [...]]",
-	"yes|no",
-	"off|dhcp|manual <address> <netmask> <gateway>",
-	"                 Remove service",
-	NULL
-};
-
-static struct option monitor_options[] = {
-	{"services", no_argument, 0, 's'},
-	{"tech", no_argument, 0, 'c'},
-	{"manager", no_argument, 0, 'm'},
+static struct connman_option monitor_options[] = {
+	{"services", 's', "[off]            Monitor only services"},
+	{"tech", 'c', "[off]            Monitor only technologies"},
+	{"manager", 'm', "[off]            Monitor only manager interface"},
+	{"vpnmanager", 'M', "[off]            Monitor only VPN manager "
+	 "interface"},
+	{"vpnconnection", 'C', "[off]            Monitor only VPN "
+	 "connections" },
 	{ NULL, }
 };
 
-static const char *monitor_desc[] = {
-	"[off]            Monitor only services",
-	"[off]            Monitor only technologies",
-	"[off]            Monitor only manager interface",
-	NULL
-};
-
 static const struct {
         const char *cmd;
 	const char *argument;
-        struct option *options;
-	const char **options_desc;
-        int (*func) (char *args[], int num, struct option *options);
+        struct connman_option *options;
+        int (*func) (char *args[], int num, struct connman_option *options);
         const char *desc;
 } cmd_table[] = {
-	{ "enable",       "<technology>|offline", NULL,    NULL,
-	  cmd_enable, "Enables given technology or offline mode" },
-	{ "disable",      "<technology>|offline", NULL,    NULL,
-	  cmd_disable, "Disables given technology or offline mode"},
-	{ "state",        NULL,           NULL,            NULL,
-	  cmd_state, "Shows if the system is online or offline" },
-	{ "services",     "[<service>]",  service_options, &service_desc[0],
-	  cmd_services, "Display services" },
-	{ "technologies", NULL,           NULL,            NULL,
-	  cmd_technologies, "Display technologies" },
-	{ "scan",         "<technology>", NULL,            NULL,
-	  cmd_scan, "Scans for new services for given technology" },
-	{ "connect",      "<service>",    NULL,            NULL,
-	  cmd_connect, "Connect a given service" },
-	{ "disconnect",   "<service>",    NULL,            NULL,
-	  cmd_disconnect, "Disconnect a given service" },
-	{ "config",       "<service>",    config_options,  &config_desc[0],
-	  cmd_config, "Set service configuration options" },
-	{ "monitor",      "[off]",        monitor_options, &monitor_desc[0],
-	  cmd_monitor, "Monitor signals from interfaces" },
-	{ "help",         NULL,           NULL,            NULL,
-	  cmd_help, "Show help" },
-	{ "exit",         NULL,           NULL,            NULL,
-	  cmd_exit,       "Exit" },
-	{ "quit",         NULL,           NULL,            NULL,
-	  cmd_exit,       "Quit" },
+	{ "state",        NULL,           NULL,            cmd_state,
+	  "Shows if the system is online or offline" },
+	{ "technologies", NULL,           NULL,            cmd_technologies,
+	  "Display technologies" },
+	{ "enable",       "<technology>|offline", NULL,    cmd_enable,
+	  "Enables given technology or offline mode" },
+	{ "disable",      "<technology>|offline", NULL,    cmd_disable,
+	  "Disables given technology or offline mode"},
+	{ "tether", "<technology> on|off\n"
+	            "            wifi [on|off] <ssid> <passphrase> ",
+	                                  NULL,            cmd_tether,
+	  "Enable, disable tethering, set SSID and passphrase for wifi" },
+	{ "services",     "[<service>]",  service_options, cmd_services,
+	  "Display services" },
+	{ "scan",         "<technology>", NULL,            cmd_scan,
+	  "Scans for new services for given technology" },
+	{ "connect",      "<service>",    NULL,            cmd_connect,
+	  "Connect a given service" },
+	{ "disconnect",   "<service>",    NULL,            cmd_disconnect,
+	  "Disconnect a given service" },
+	{ "config",       "<service>",    config_options,  cmd_config,
+	  "Set service configuration options" },
+	{ "monitor",      "[off]",        monitor_options, cmd_monitor,
+	  "Monitor signals from interfaces" },
+	{ "agent", "on|off",              NULL,            cmd_agent,
+	  "Agent mode" },
+	{"vpnconnections", "[<connection>]", NULL,         cmd_vpnconnections,
+	 "Display VPN connections" },
+	{ "vpnagent",     "on|off",     NULL,            cmd_vpnagent,
+	  "VPN Agent mode" },
+	{ "help",         NULL,           NULL,            cmd_help,
+	  "Show help" },
+	{ "exit",         NULL,           NULL,            cmd_exit,
+	  "Exit" },
+	{ "quit",         NULL,           NULL,            cmd_exit,
+	  "Quit" },
 	{  NULL, },
 };
 
-static int cmd_help(char *args[], int num, struct option *options)
+static int cmd_help(char *args[], int num, struct connman_option *options)
 {
 	bool interactive = __connmanctl_is_interactive();
 	int i, j;
@@ -1056,7 +1466,7 @@ static int cmd_help(char *args[], int num, struct option *options)
 		const char *argument = cmd_table[i].argument;
 		const char *desc = cmd_table[i].desc;
 
-		printf("%-12s%-22s%s\n", cmd != NULL? cmd: "",
+		printf("%-16s%-22s%s\n", cmd != NULL? cmd: "",
 				argument != NULL? argument: "",
 				desc != NULL? desc: "");
 
@@ -1064,10 +1474,10 @@ static int cmd_help(char *args[], int num, struct option *options)
 			for (j = 0; cmd_table[i].options[j].name != NULL;
 			     j++) {
 				const char *options_desc =
-					cmd_table[i].options_desc != NULL ?
-					cmd_table[i].options_desc[j]: "";
+					cmd_table[i].options[j].desc != NULL ?
+					cmd_table[i].options[j].desc: "";
 
-				printf("   --%-12s%s\n",
+				printf("   --%-16s%s\n",
 						cmd_table[i].options[j].name,
 						options_desc);
 			}
@@ -1081,7 +1491,7 @@ static int cmd_help(char *args[], int num, struct option *options)
 	return 0;
 }
 
-int commands(DBusConnection *dbus_conn, char *argv[], int argc)
+int __connmanctl_commands(DBusConnection *dbus_conn, char *argv[], int argc)
 {
 	int i, result;
 
@@ -1102,3 +1512,25 @@ int commands(DBusConnection *dbus_conn, char *argv[], int argc)
 	fprintf(stderr, "Error '%s': Unknown command\n", argv[0]);
 	return -EINVAL;
 }
+
+char *__connmanctl_lookup_command(const char *text, int state)
+{
+	static int i = 0;
+	static int len = 0;
+
+	if (state == 0) {
+		i = 0;
+		len = strlen(text);
+	}
+
+	while (cmd_table[i].cmd != NULL) {
+		const char *command = cmd_table[i].cmd;
+
+		i++;
+
+		if (strncmp(text, command, len) == 0)
+			return strdup(command);
+	}
+
+	return NULL;
+}
