@@ -172,8 +172,8 @@ uint8_t *dhcpv6_get_option(struct dhcpv6_packet *packet, uint16_t pkt_len,
 		if (opt_code == code) {
 			if (option_len != NULL)
 				*option_len = opt_len;
-			if (rem == 0)
-				found = NULL;
+			if (rem < 0)
+				goto bad_packet;
 			else
 				found = optionptr + 2 + 2;
 			count++;
@@ -525,8 +525,17 @@ int dhcpv6_send_packet(int index, struct dhcpv6_packet *dhcp_pkt, int len)
 	m.msg_controllen = cmsg->cmsg_len;
 
 	ret = sendmsg(fd, &m, 0);
-	if (ret < 0)
-		perror("DHCPv6 msg send failed");
+	if (ret < 0) {
+		char *msg = "DHCPv6 msg send failed";
+
+		if (errno == EADDRNOTAVAIL) {
+			char *str = g_strdup_printf("%s (index %d)",
+					msg, index);
+			perror(str);
+			g_free(str);
+		} else
+			perror(msg);
+	}
 
 	g_free(control_buf);
 	close(fd);
@@ -629,7 +638,7 @@ int dhcp_send_kernel_packet(struct dhcp_packet *dhcp_pkt,
 	memset(&client, 0, sizeof(client));
 	client.sin_family = AF_INET;
 	client.sin_port = htons(dest_port);
-	client.sin_addr.s_addr = dest_ip;
+	client.sin_addr.s_addr = htonl(dest_ip);
 	if (connect(fd, (struct sockaddr *) &client, sizeof(client)) < 0) {
 		close(fd);
 		return -errno;
