@@ -106,10 +106,7 @@ static struct gateway_data *lookup_gateway_data(struct gateway_config *config)
 	return NULL;
 }
 
-/*
- * Find the gateway that is serving the VPN link
- */
-static struct gateway_data *find_phy_gateway(int index, const char *gateway)
+static struct gateway_data *find_vpn_gateway(int index, const char *gateway)
 {
 	GHashTableIter iter;
 	gpointer value, key;
@@ -122,12 +119,12 @@ static struct gateway_data *find_phy_gateway(int index, const char *gateway)
 	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
 		struct gateway_data *data = value;
 
-		if (data->ipv4_gateway != NULL && data->index != index &&
+		if (data->ipv4_gateway != NULL && data->index == index &&
 				g_str_equal(data->ipv4_gateway->gateway,
 					gateway) == TRUE)
 			return data;
 
-		if (data->ipv6_gateway != NULL && data->index != index &&
+		if (data->ipv6_gateway != NULL && data->index == index &&
 				g_str_equal(data->ipv6_gateway->gateway,
 					gateway) == TRUE)
 			return data;
@@ -136,58 +133,135 @@ static struct gateway_data *find_phy_gateway(int index, const char *gateway)
 	return NULL;
 }
 
-static void set_vpn_routes(struct gateway_config *config,
+struct get_gateway_params {
+	char *vpn_gateway;
+	int vpn_index;
+};
+
+static void get_gateway_cb(const char *gateway, int index, void *user_data)
+{
+	struct gateway_config *config;
+	struct gateway_data *data;
+	struct get_gateway_params *params = user_data;
+	int family;
+
+	if (index < 0)
+		goto out;
+
+	DBG("phy index %d phy gw %s vpn index %d vpn gw %s", index, gateway,
+		params->vpn_index, params->vpn_gateway);
+
+	data = find_vpn_gateway(params->vpn_index, params->vpn_gateway);
+	if (data == NULL) {
+		DBG("Cannot find VPN link route, index %d addr %s",
+			params->vpn_index, params->vpn_gateway);
+		goto out;
+	}
+
+	family = connman_inet_check_ipaddress(params->vpn_gateway);
+
+	if (family == AF_INET)
+		config = data->ipv4_gateway;
+	else if (family == AF_INET6)
+		config = data->ipv6_gateway;
+	else
+		goto out;
+
+	config->vpn_phy_index = index;
+
+	DBG("vpn %s phy index %d", config->vpn_ip, config->vpn_phy_index);
+
+out:
+	g_free(params->vpn_gateway);
+	g_free(params);
+}
+
+static void set_vpn_routes(struct gateway_data *new_gateway,
 			struct connman_service *service,
 			const char *gateway,
 			enum connman_ipconfig_type type,
-			const char *peer)
+			const char *peer,
+			struct gateway_data *active_gateway)
 {
-	struct gateway_data *data;
+	struct gateway_config *config;
 	struct connman_ipconfig *ipconfig;
-	int index;
+	char *dest;
 
-	config->vpn = TRUE;
-	if (peer != NULL)
-		config->vpn_ip = g_strdup(peer);
-	else if (gateway != NULL)
-		config->vpn_ip = g_strdup(gateway);
+	DBG("new %p service %p gw %s type %d peer %s active %p",
+		new_gateway, service, gateway, type, peer, active_gateway);
 
-	if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
+	if (type == CONNMAN_IPCONFIG_TYPE_IPV4) {
 		ipconfig = __connman_service_get_ip4config(service);
-	else if (type == CONNMAN_IPCONFIG_TYPE_IPV6)
+		config = new_gateway->ipv4_gateway;
+	} else if (type == CONNMAN_IPCONFIG_TYPE_IPV6) {
 		ipconfig = __connman_service_get_ip6config(service);
-	else
+		config = new_gateway->ipv6_gateway;
+	} else
 		return;
 
-	index = __connman_ipconfig_get_index(ipconfig);
-	data = find_phy_gateway(index, gateway);
+	if (config != NULL) {
+		int index = __connman_ipconfig_get_index(ipconfig);
+		struct get_gateway_params *params;
 
-	if (data != NULL) {
-		/*
-		 * data->service points now to original
-		 * service that is serving the VPN link
-		 */
-		if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
-			ipconfig =
-				__connman_service_get_ip4config(data->service);
-		else if (type == CONNMAN_IPCONFIG_TYPE_IPV6)
-			ipconfig =
-				__connman_service_get_ip6config(data->service);
-		else
+		config->vpn = TRUE;
+		if (peer != NULL)
+			config->vpn_ip = g_strdup(peer);
+		else if (gateway != NULL)
+			config->vpn_ip = g_strdup(gateway);
+
+		params = g_try_malloc(sizeof(struct get_gateway_params));
+		if (params == NULL)
 			return;
 
-		if (ipconfig != NULL) {
-			const char *address;
+		params->vpn_index = index;
+		params->vpn_gateway = g_strdup(gateway);
 
-			address = __connman_ipconfig_get_local(ipconfig);
-			config->vpn_phy_ip = g_strdup(address);
-		}
-
-		config->vpn_phy_index = data->index;
+		/*
+		 * Find the gateway that is serving the VPN link
+		 */
+		__connman_inet_get_route(gateway, get_gateway_cb, params);
 	}
 
-	DBG("vpn %s phy %s index %d", config->vpn_ip,
-		config->vpn_phy_ip, config->vpn_phy_index);
+	if (active_gateway == NULL)
+		return;
+
+	if (type == CONNMAN_IPCONFIG_TYPE_IPV4) {
+		/*
+		 * Special route to VPN server via gateway. This
+		 * is needed so that we can access hosts behind
+		 * the VPN. The route might already exist depending
+		 * on network topology.
+		 */
+		if (active_gateway->ipv4_gateway == NULL)
+			return;
+
+		DBG("active gw %s", active_gateway->ipv4_gateway->gateway);
+
+		if (g_strcmp0(active_gateway->ipv4_gateway->gateway,
+							"0.0.0.0") != 0)
+			dest = active_gateway->ipv4_gateway->gateway;
+		else
+			dest = NULL;
+
+		connman_inet_add_host_route(active_gateway->index, gateway,
+									dest);
+
+	} else if (type == CONNMAN_IPCONFIG_TYPE_IPV6) {
+
+		if (active_gateway->ipv6_gateway == NULL)
+			return;
+
+		DBG("active gw %s", active_gateway->ipv6_gateway->gateway);
+
+		if (g_strcmp0(active_gateway->ipv6_gateway->gateway,
+								"::") != 0)
+			dest = active_gateway->ipv6_gateway->gateway;
+		else
+			dest = NULL;
+
+		connman_inet_add_ipv6_host_route(active_gateway->index,
+								gateway, dest);
+	}
 }
 
 static int del_routes(struct gateway_data *data,
@@ -358,10 +432,7 @@ static void set_default_gateway(struct gateway_data *data,
 
 	if (do_ipv4 == TRUE && data->ipv4_gateway != NULL &&
 					data->ipv4_gateway->vpn == TRUE) {
-		connman_inet_set_gateway_address(data->index,
-						data->ipv4_gateway->vpn_ip);
-		connman_inet_add_host_route(data->index,
-					data->ipv4_gateway->vpn_ip, NULL);
+		connman_inet_set_gateway_interface(data->index);
 		data->ipv4_gateway->active = TRUE;
 
 		DBG("set %p index %d vpn %s index %d phy %s",
@@ -376,10 +447,7 @@ static void set_default_gateway(struct gateway_data *data,
 
 	if (do_ipv6 == TRUE && data->ipv6_gateway != NULL &&
 					data->ipv6_gateway->vpn == TRUE) {
-		connman_inet_set_ipv6_gateway_address(data->index,
-						data->ipv6_gateway->vpn_ip);
-		connman_inet_add_ipv6_host_route(data->index,
-					data->ipv6_gateway->vpn_ip, NULL);
+		connman_inet_set_ipv6_gateway_interface(data->index);
 		data->ipv6_gateway->active = TRUE;
 
 		DBG("set %p index %d vpn %s index %d phy %s",
@@ -411,12 +479,12 @@ static void set_default_gateway(struct gateway_data *data,
 	}
 
 	if (do_ipv6 == TRUE && data->ipv6_gateway != NULL)
-		status6 = connman_inet_set_ipv6_gateway_address(index,
-						data->ipv6_gateway->gateway);
+		status6 = __connman_inet_add_default_to_table(RT_TABLE_MAIN,
+					index, data->ipv6_gateway->gateway);
 
 	if (do_ipv4 == TRUE && data->ipv4_gateway != NULL)
-		status4 = connman_inet_set_gateway_address(index,
-						data->ipv4_gateway->gateway);
+		status4 = __connman_inet_add_default_to_table(RT_TABLE_MAIN,
+					index, data->ipv4_gateway->gateway);
 
 	if (status4 < 0 || status6 < 0)
 		return;
@@ -443,10 +511,7 @@ static void unset_default_gateway(struct gateway_data *data,
 
 	if (do_ipv4 == TRUE && data->ipv4_gateway != NULL &&
 					data->ipv4_gateway->vpn == TRUE) {
-		connman_inet_del_host_route(data->index,
-						data->ipv4_gateway->vpn_ip);
-		connman_inet_clear_gateway_address(data->index,
-						data->ipv4_gateway->vpn_ip);
+		connman_inet_clear_gateway_interface(data->index);
 		data->ipv4_gateway->active = FALSE;
 
 		DBG("unset %p index %d vpn %s index %d phy %s",
@@ -459,10 +524,7 @@ static void unset_default_gateway(struct gateway_data *data,
 
 	if (do_ipv6 == TRUE && data->ipv6_gateway != NULL &&
 					data->ipv6_gateway->vpn == TRUE) {
-		connman_inet_del_ipv6_host_route(data->index,
-						data->ipv6_gateway->vpn_ip);
-		connman_inet_clear_ipv6_gateway_address(data->index,
-						data->ipv6_gateway->vpn_ip);
+		connman_inet_clear_ipv6_gateway_interface(data->index);
 		data->ipv6_gateway->active = FALSE;
 
 		DBG("unset %p index %d vpn %s index %d phy %s",
@@ -833,30 +895,8 @@ int __connman_connection_gateway_add(struct connman_service *service,
 
 	if (service_type == CONNMAN_SERVICE_TYPE_VPN) {
 
-		if (type == CONNMAN_IPCONFIG_TYPE_IPV4) {
-			if (new_gateway->ipv4_gateway != NULL)
-				set_vpn_routes(new_gateway->ipv4_gateway,
-					service, gateway, type, peer);
-
-			/*
-			 * Special route to VPN server via gateway. This
-			 * is needed so that we can access hosts behind
-			 * the VPN. The route might already exist depending
-			 * on network topology.
-			 */
-			connman_inet_add_host_route(active_gateway->index,
-					gateway,
-					active_gateway->ipv4_gateway->gateway);
-
-		} else if (type == CONNMAN_IPCONFIG_TYPE_IPV6) {
-			if (new_gateway->ipv6_gateway != NULL)
-				set_vpn_routes(new_gateway->ipv6_gateway,
-					service, gateway, type, peer);
-
-			connman_inet_add_ipv6_host_route(active_gateway->index,
-					gateway,
-					active_gateway->ipv6_gateway->gateway);
-		}
+		set_vpn_routes(new_gateway, service, gateway, type, peer,
+							active_gateway);
 
 	} else {
 		if (type == CONNMAN_IPCONFIG_TYPE_IPV4 &&
