@@ -56,6 +56,13 @@ DBusMessage *__connman_error_invalid_property(DBusMessage *msg);
 int __connman_manager_init(void);
 void __connman_manager_cleanup(void);
 
+enum time_updates {
+	TIME_UPDATES_UNKNOWN = 0,
+	TIME_UPDATES_MANUAL  = 1,
+	TIME_UPDATES_AUTO    = 2,
+};
+
+enum time_updates __connman_clock_timeupdates(void);
 int __connman_clock_init(void);
 void __connman_clock_cleanup(void);
 
@@ -133,6 +140,8 @@ int __connman_inet_modify_address(int cmd, int flags, int index, int family,
 				const char *peer,
 				unsigned char prefixlen,
 				const char *broadcast);
+int __connman_inet_get_interface_address(int index, int family, void *address);
+int __connman_inet_get_interface_ll_address(int index, int family, void *address);
 
 #include <netinet/ip6.h>
 #include <netinet/icmp6.h>
@@ -142,8 +151,17 @@ typedef void (*__connman_inet_rs_cb_t) (struct nd_router_advert *reply,
 
 int __connman_inet_ipv6_send_rs(int index, int timeout,
 			__connman_inet_rs_cb_t callback, void *user_data);
+int __connman_inet_ipv6_send_ra(int index, struct in6_addr *src_addr,
+				GSList *prefixes, int router_lifetime);
 
-int __connman_refresh_rs_ipv6(struct connman_network *network, int index);
+typedef void (*__connman_inet_recv_rs_cb_t) (struct nd_router_solicit *reply,
+					unsigned int length, void *user_data);
+int __connman_inet_ipv6_start_recv_rs(int index,
+				__connman_inet_recv_rs_cb_t callback,
+				void *user_data, void **context);
+void __connman_inet_ipv6_stop_recv_rs(void *context);
+
+int __connman_network_refresh_rs_ipv6(struct connman_network *network, int index);
 
 GSList *__connman_inet_ipv6_get_prefixes(struct nd_router_advert *hdr,
 					unsigned int length);
@@ -192,6 +210,11 @@ int __connman_inet_rtnl_addattr_l(struct nlmsghdr *n, size_t max_length,
 int __connman_inet_rtnl_addattr32(struct nlmsghdr *n, size_t maxlen,
 			int type, __u32 data);
 
+int __connman_inet_add_fwmark_rule(uint32_t table_id, int family, uint32_t fwmark);
+int __connman_inet_del_fwmark_rule(uint32_t table_id, int family, uint32_t fwmark);
+int __connman_inet_add_default_to_table(uint32_t table_id, int ifindex, const char *gateway);
+int __connman_inet_del_default_from_table(uint32_t table_id, int ifindex, const char *gateway);
+
 #include <connman/resolver.h>
 
 int __connman_resolver_init(connman_bool_t dnsproxy);
@@ -356,6 +379,9 @@ const char *__connman_ipconfig_get_proxy_autoconfig(struct connman_ipconfig *ipc
 void __connman_ipconfig_set_dhcp_address(struct connman_ipconfig *ipconfig,
 					const char *address);
 char *__connman_ipconfig_get_dhcp_address(struct connman_ipconfig *ipconfig);
+void __connman_ipconfig_set_dhcpv6_prefixes(struct connman_ipconfig *ipconfig,
+					char **prefixes);
+char **__connman_ipconfig_get_dhcpv6_prefixes(struct connman_ipconfig *ipconfig);
 
 int __connman_ipconfig_load(struct connman_ipconfig *ipconfig,
 		GKeyFile *keyfile, const char *identifier, const char *prefix);
@@ -364,6 +390,7 @@ int __connman_ipconfig_save(struct connman_ipconfig *ipconfig,
 gboolean __connman_ipconfig_ipv6_privacy_enabled(struct connman_ipconfig *ipconfig);
 int __connman_ipconfig_ipv6_set_privacy(struct connman_ipconfig *ipconfig,
 					const char *value);
+gboolean __connman_ipconfig_ipv6_is_enabled(struct connman_ipconfig *ipconfig);
 
 int __connman_ipconfig_set_rp_filter();
 void __connman_ipconfig_unset_rp_filter(int old_value);
@@ -387,7 +414,7 @@ int __connman_timeserver_sync(struct connman_service *service);
 void __connman_timeserver_sync_next();
 
 typedef void (* dhcp_cb) (struct connman_network *network,
-				connman_bool_t success);
+			connman_bool_t success, gpointer data);
 int __connman_dhcp_start(struct connman_network *network, dhcp_cb callback);
 void __connman_dhcp_stop(struct connman_network *network);
 int __connman_dhcp_init(void);
@@ -403,6 +430,12 @@ int __connman_dhcpv6_start_renew(struct connman_network *network,
 				dhcp_cb callback);
 int __connman_dhcpv6_start_release(struct connman_network *network,
 				dhcp_cb callback);
+int __connman_dhcpv6_start_pd(int index, GSList *prefixes, dhcp_cb callback);
+void __connman_dhcpv6_stop_pd(int index);
+int __connman_dhcpv6_start_pd_renew(struct connman_network *network,
+							dhcp_cb callback);
+int __connman_dhcpv6_start_pd_release(struct connman_network *network,
+				dhcp_cb callback);
 
 int __connman_ipv4_init(void);
 void __connman_ipv4_cleanup(void);
@@ -545,6 +578,9 @@ void __connman_tethering_set_disabled(void);
 int __connman_private_network_request(DBusMessage *msg, const char *owner);
 int __connman_private_network_release(const char *path);
 
+int __connman_ipv6pd_setup(const char *bridge);
+void __connman_ipv6pd_cleanup(void);
+
 #include <connman/provider.h>
 
 connman_bool_t __connman_provider_check_routes(struct connman_provider *provider);
@@ -559,7 +595,7 @@ int __connman_provider_indicate_state(struct connman_provider *provider,
 int __connman_provider_indicate_error(struct connman_provider *provider,
 					enum connman_provider_error error);
 int __connman_provider_connect(struct connman_provider *provider);
-int __connman_provider_remove(const char *path);
+int __connman_provider_remove_by_path(const char *path);
 void __connman_provider_cleanup(void);
 int __connman_provider_init(void);
 
@@ -609,6 +645,8 @@ void __connman_service_set_userconnect(struct connman_service *service,
 						connman_bool_t userconnect);
 void __connman_service_set_search_domains(struct connman_service *service,
 					char **domains);
+void __connman_service_update_search_domains(struct connman_service *service,
+					char **domains);
 
 void __connman_service_set_string(struct connman_service *service,
 					const char *key, const char *value);
@@ -702,18 +740,12 @@ void __connman_service_counter_unregister(const char *counter);
 
 #include <connman/session.h>
 
-struct service_entry;
-typedef connman_bool_t (* service_match_cb) (struct connman_session *session,
-					struct connman_service *service);
-typedef struct service_entry* (* create_service_entry_cb) (
-					struct connman_service *service,
+typedef void (* service_iterate_cb) (struct connman_service *service,
 					const char *name,
-					enum connman_service_state state);
+					enum connman_service_state state,
+					void *user_data);
 
-GSequence *__connman_service_get_list(struct connman_session *session,
-				service_match_cb service_match,
-				create_service_entry_cb create_service_entry,
-				GDestroyNotify destroy_service_entry);
+int __connman_service_iterate_services(service_iterate_cb cb, void *user_data);
 
 void __connman_service_session_inc(struct connman_service *service);
 connman_bool_t __connman_service_session_dec(struct connman_service *service);
@@ -790,6 +822,7 @@ int __connman_stats_get(struct connman_service *service,
 				connman_bool_t roaming,
 				struct connman_stats_data *data);
 
+int __connman_iptables_dump(const char *table_name);
 int __connman_iptables_new_chain(const char *table_name,
 					const char *chain);
 int __connman_iptables_delete_chain(const char *table_name,
@@ -802,10 +835,19 @@ int __connman_iptables_change_policy(const char *table_name,
 int __connman_iptables_append(const char *table_name,
 			const char *chain,
 			const char *rule_spec);
+int __connman_iptables_insert(const char *table_name,
+			const char *chain,
+			const char *rule_spec);
 int __connman_iptables_delete(const char *table_name,
 			const char *chain,
 			const char *rule_spec);
 
+typedef void (*connman_iptables_iterate_chains_cb_t) (const char *chain_name,
+							void *user_data);
+int __connman_iptables_iterate_chains(const char *table_name,
+				connman_iptables_iterate_chains_cb_t cb,
+				void *user_data);
+
 int __connman_iptables_init(void);
 void __connman_iptables_cleanup(void);
 int __connman_iptables_commit(const char *table_name);
@@ -859,8 +901,8 @@ void __connman_ippool_deladdr(int index, const char *address,
 
 int __connman_bridge_create(const char *name);
 int __connman_bridge_remove(const char *name);
-int __connman_bridge_enable(const char *name, const char *gateway,
-				const char *broadcast);
+int __connman_bridge_enable(const char *name, const char *ip_address,
+			int prefix_len, const char *broadcast);
 int __connman_bridge_disable(const char *name);
 
 int __connman_nat_init(void);
@@ -869,3 +911,51 @@ void __connman_nat_cleanup(void);
 int __connman_nat_enable(const char *name, const char *address,
 				unsigned char prefixlen);
 void __connman_nat_disable(const char *name);
+
+struct firewall_context;
+
+struct firewall_context *__connman_firewall_create(void);
+void __connman_firewall_destroy(struct firewall_context *ctx);
+int __connman_firewall_add_rule(struct firewall_context *ctx,
+				const char *table,
+				const char *chain,
+				const char *rule_fmt, ...);
+int __connman_firewall_enable(struct firewall_context *ctx);
+int __connman_firewall_disable(struct firewall_context *ctx);
+
+int __connman_firewall_init(void);
+void __connman_firewall_cleanup(void);
+
+typedef int (* connman_nfacct_flush_cb_t) (unsigned int error, void *user_data);
+
+int __connman_nfacct_flush(connman_nfacct_flush_cb_t cb, void *user_data);
+
+struct nfacct_context;
+
+typedef void (* connman_nfacct_enable_cb_t) (unsigned int error,
+						struct nfacct_context *ctx,
+						void *user_data);
+typedef void (* connman_nfacct_disable_cb_t) (unsigned int error,
+						struct nfacct_context *ctx,
+						void *user_data);
+typedef void (* connman_nfacct_stats_cb_t) (struct nfacct_context *ctx,
+						uint64_t packets,
+						uint64_t bytes,
+						void *user_data);
+
+struct nfacct_context *__connman_nfacct_create_context(void);
+void __connman_nfacct_destroy_context(struct nfacct_context *ctx);
+
+int __connman_nfacct_add(struct nfacct_context *ctx, const char *name,
+				connman_nfacct_stats_cb_t cb,
+				void *user_data);
+int __connman_nfacct_enable(struct nfacct_context *ctx,
+				connman_nfacct_enable_cb_t cb,
+				void *user_data);
+int __connman_nfacct_disable(struct nfacct_context *ctx,
+				connman_nfacct_disable_cb_t cb,
+				void *user_data);
+
+
+int __connman_nfacct_init(void);
+void __connman_nfacct_cleanup(void);
