@@ -122,7 +122,7 @@ static void dhcp_invalidate(struct connman_dhcp *dhcp, connman_bool_t callback)
 	__connman_ipconfig_set_prefixlen(ipconfig, 0);
 
 	if (dhcp->callback != NULL && callback)
-		dhcp->callback(dhcp->network, FALSE);
+		dhcp->callback(dhcp->network, FALSE, NULL);
 
 out:
 	dhcp_free(dhcp);
@@ -131,7 +131,7 @@ out:
 static void dhcp_valid(struct connman_dhcp *dhcp)
 {
 	if (dhcp->callback != NULL)
-		dhcp->callback(dhcp->network, TRUE);
+		dhcp->callback(dhcp->network, TRUE, NULL);
 }
 
 static void no_lease_cb(GDHCPClient *dhcp_client, gpointer user_data)
@@ -227,7 +227,7 @@ static void lease_available_cb(GDHCPClient *dhcp_client, gpointer user_data)
 	if (option != NULL)
 		gateway = g_strdup(option->data);
 
-	prefixlen = __connman_ipconfig_netmask_prefix_len(netmask);
+	prefixlen = __connman_ipaddress_netmask_prefix_len(netmask);
 	if (prefixlen == 255)
 		connman_warn("netmask: %s is invalid", netmask);
 
@@ -375,7 +375,7 @@ static void ipv4ll_available_cb(GDHCPClient *dhcp_client, gpointer user_data)
 	address = g_dhcp_client_get_address(dhcp_client);
 	netmask = g_dhcp_client_get_netmask(dhcp_client);
 
-	prefixlen = __connman_ipconfig_netmask_prefix_len(netmask);
+	prefixlen = __connman_ipaddress_netmask_prefix_len(netmask);
 
 	__connman_ipconfig_set_method(ipconfig, CONNMAN_IPCONFIG_METHOD_DHCP);
 	__connman_ipconfig_set_local(ipconfig, address);
@@ -449,6 +449,12 @@ static int dhcp_request(struct connman_dhcp *dhcp)
 	service = connman_service_lookup_from_network(dhcp->network);
 	ipconfig = __connman_service_get_ip4config(service);
 
+	/*
+	 * Clear the addresses at startup so that lease callback will
+	 * take the lease and set ip address properly.
+	 */
+	__connman_ipconfig_clear_address(ipconfig);
+
 	return g_dhcp_client_start(dhcp_client,
 				__connman_ipconfig_get_dhcp_address(ipconfig));
 }
