@@ -70,6 +70,8 @@ static struct {
 	char **blacklisted_interfaces;
 	connman_bool_t allow_hostname_updates;
 	connman_bool_t single_tech;
+	char **tethering_technologies;
+	connman_bool_t persistent_tethering_mode;
 } connman_settings  = {
 	.bg_scan = TRUE,
 	.pref_timeservers = NULL,
@@ -81,6 +83,8 @@ static struct {
 	.blacklisted_interfaces = NULL,
 	.allow_hostname_updates = TRUE,
 	.single_tech = FALSE,
+	.tethering_technologies = NULL,
+	.persistent_tethering_mode = FALSE,
 };
 
 #define CONF_BG_SCAN                    "BackgroundScanning"
@@ -93,6 +97,8 @@ static struct {
 #define CONF_BLACKLISTED_INTERFACES     "NetworkInterfaceBlacklist"
 #define CONF_ALLOW_HOSTNAME_UPDATES     "AllowHostnameUpdates"
 #define CONF_SINGLE_TECH                "SingleConnectedTechnology"
+#define CONF_TETHERING_TECHNOLOGIES      "TetheringTechnologies"
+#define CONF_PERSISTENT_TETHERING_MODE  "PersistentTetheringMode"
 
 static const char *supported_options[] = {
 	CONF_BG_SCAN,
@@ -105,6 +111,8 @@ static const char *supported_options[] = {
 	CONF_BLACKLISTED_INTERFACES,
 	CONF_ALLOW_HOSTNAME_UPDATES,
 	CONF_SINGLE_TECH,
+	CONF_TETHERING_TECHNOLOGIES,
+	CONF_PERSISTENT_TETHERING_MODE,
 	NULL
 };
 
@@ -225,6 +233,7 @@ static void parse_config(GKeyFile *config)
 	char **timeservers;
 	char **interfaces;
 	char **str_list;
+	char **tethering;
 	gsize len;
 	int timeout;
 
@@ -327,6 +336,22 @@ static void parse_config(GKeyFile *config)
 		connman_settings.single_tech = boolean;
 
 	g_clear_error(&error);
+
+	tethering = g_key_file_get_string_list(config, "General",
+			CONF_TETHERING_TECHNOLOGIES, &len, &error);
+
+	if (error == NULL)
+		connman_settings.tethering_technologies = tethering;
+
+	g_clear_error(&error);
+
+	boolean = g_key_file_get_boolean(config, "General",
+					CONF_PERSISTENT_TETHERING_MODE,
+					&error);
+	if (error == NULL)
+		connman_settings.persistent_tethering_mode = boolean;
+
+	g_clear_error(&error);
 }
 
 static int config_init(const char *file)
@@ -498,6 +523,9 @@ connman_bool_t connman_setting_get_bool(const char *key)
 	if (g_str_equal(key, CONF_SINGLE_TECH) == TRUE)
 		return connman_settings.single_tech;
 
+	if (g_str_equal(key, CONF_PERSISTENT_TETHERING_MODE) == TRUE)
+		return connman_settings.persistent_tethering_mode;
+
 	return FALSE;
 }
 
@@ -512,6 +540,9 @@ char **connman_setting_get_string_list(const char *key)
 	if (g_str_equal(key, CONF_BLACKLISTED_INTERFACES) == TRUE)
 		return connman_settings.blacklisted_interfaces;
 
+	if (g_str_equal(key, CONF_TETHERING_TECHNOLOGIES) == TRUE)
+		return connman_settings.tethering_technologies;
+
 	return NULL;
 }
 
@@ -542,11 +573,6 @@ int main(int argc, char *argv[])
 	DBusError err;
 	guint signal;
 
-#ifdef NEED_THREADS
-	if (g_thread_supported() == FALSE)
-		g_thread_init(NULL);
-#endif
-
 	context = g_option_context_new(NULL);
 	g_option_context_add_main_entries(context, options, NULL);
 
@@ -589,13 +615,6 @@ int main(int argc, char *argv[])
 
 	main_loop = g_main_loop_new(NULL, FALSE);
 
-#ifdef NEED_THREADS
-	if (dbus_threads_init_default() == FALSE) {
-		fprintf(stderr, "Can't init usage of threads\n");
-		exit(1);
-	}
-#endif
-
 	signal = setup_signalfd();
 
 	dbus_error_init(&err);
@@ -633,6 +652,8 @@ int main(int argc, char *argv[])
 
 	__connman_ippool_init();
 	__connman_iptables_init();
+	__connman_nfacct_init();
+	__connman_firewall_init();
 	__connman_nat_init();
 	__connman_tethering_init();
 	__connman_counter_init();
@@ -675,11 +696,11 @@ int main(int argc, char *argv[])
 	__connman_wpad_cleanup();
 	__connman_dhcpv6_cleanup();
 	__connman_dhcp_cleanup();
+	__connman_session_cleanup();
 	__connman_plugin_cleanup();
 	__connman_provider_cleanup();
 	__connman_connection_cleanup();
 	__connman_timeserver_cleanup();
-	__connman_session_cleanup();
 	__connman_detect_cleanup();
 	__connman_proxy_cleanup();
 	__connman_task_cleanup();
@@ -693,6 +714,8 @@ int main(int argc, char *argv[])
 	__connman_counter_cleanup();
 	__connman_tethering_cleanup();
 	__connman_nat_cleanup();
+	__connman_firewall_cleanup();
+	__connman_nfacct_cleanup();
 	__connman_iptables_cleanup();
 	__connman_ippool_cleanup();
 	__connman_device_cleanup();
@@ -719,6 +742,7 @@ int main(int argc, char *argv[])
 	g_free(connman_settings.preferred_techs);
 	g_strfreev(connman_settings.fallback_nameservers);
 	g_strfreev(connman_settings.blacklisted_interfaces);
+	g_strfreev(connman_settings.tethering_technologies);
 
 	g_free(option_debug);
 
